\\ Copyright 2014 Kevin Ryde
\\
\\ memoize.gp is free software; you can redistribute it and/or modify it
\\ under the terms of the GNU General Public License as published by the Free
\\ Software Foundation; either version 3, or (at your option) any later
\\ version.
\\
\\ memoize.gp is distributed in the hope that it will be useful, but
\\ WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
\\ or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
\\ for more details.
\\
\\ You should have received a copy of the GNU General Public License along
\\ with memoize.gp  See the file COPYING.  If not, see
\\ <http://www.gnu.org/licenses/>.

{
my(bracket=Vecsmall("[")[1]);
my(open=Vecsmall("(")[1]);
my(close=Vecsmall(")")[1]);
my(comma=Vecsmall(",")[1]);
memoize__arity(func) =
  type(func) == "t_CLOSURE" || error("not a function");
  my(v=Vecsmall(Str(func)));
  if(v[2]==close, return(0));  \\ () no args
  my(arity=1, nest=0);
  \\ count commas up to closing ")"
  for(i=2,#v,
     if(v[i]==open,    nest++,
        v[i]==close,   nest--; if(nest<0,return(arity)),
        v[i]==bracket, return(-arity),  \\ varargs
        v[i]==comma,   arity++));       \\ count commas
  error("memoize__arity() oops, no closing paren");
}
{
addhelp(memoize__arity,
"n = memoize__arity(func)
Return the number of arguments taken by function func.
This is an internal part of memoize.gp.")
}






zero_args()="ret zero"
test_equal(memoize__arity(zero_args), 0)

one_args(n)="ret one"
test_equal(memoize__arity(one_args), 1)

two_args(abc0,def1)="ret two"
test_equal(memoize__arity(two_args), 2)

default_expr(x=(1+(2+3)+4),y) = x+y
test_equal(memoize__arity(default_expr), 2)

\\ strings as defaults not recognised by memoize__arity()
\\ default_str(x="(") = x+y
\\ test_equal(memoize__arity(default_str), 1)




memoize__func_args_full(func) =
  type(func) == "t_CLOSURE" || error("not a function");
  my(v = Vecsmall(Str(func)),
     nest = 0);
  for(i=2,#v,
     if(setsearch(open_set,v[i]),   nest++; next,
        setsearch(close_set,v[i]),  nest--; if(nest<0,
                                               return(if(i==2,"",
                                                         Strchr(v[2..i-1]))))));
  error("memoize__func_args_full() oops, no closing paren")
);
addhelp(memoize__func_args_full,
"str = memoize__func_args_full(func)
This is an internal part of memoize.gp.

Return a string which is the function arguments of function func.
This is everything between the parens which are the function formal
parameters, including any default values expressions.

For example foo(x,y=2) = 3*x+4*y returns \"x,y=2\".
");
zero_args()="ret zero"
test_equal(memoize__func_argnames(zero_args), "")
test_equal(memoize__func_args_full(zero_args), "")
test_equal(memoize__func_is_zero_args(zero_args), 1)

one_args(n)="ret one"
test_equal(memoize__func_argnames(one_args), "n")
test_equal(memoize__func_args_full(one_args), "n")
test_equal(memoize__func_is_zero_args(one_args), 0)

two_args(abc0,def1)="ret two"
test_equal(memoize__func_argnames(two_args), "abc0,def1")
test_equal(memoize__func_args_full(two_args), "abc0,def1")
test_equal(memoize__func_is_zero_args(two_args), 0)

default_expr(x=(1+(2+3)+4),y) = x+y
test_equal(memoize__func_argnames(default_expr), "x,y")
test_equal(memoize__func_args_full(default_expr), "x=(1+(2+3)+4),y")
test_equal(memoize__func_is_zero_args(default_expr), 0)

